/* ltc2amidi.c - LTC/SMPTE to MTC encoder
 * (c) 2006, 2009 Robin Gareus <robin@gareus.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <ltcsmpte.h>

#include <alsa/asoundlib.h>
#include <pthread.h>

extern int want_verbose;

snd_seq_t *seq= NULL;
int portnum;

/*
 *  Structure of a System Exclusive Message
 *
 *  # Start of Exclusive: F0
 *  # Manufacturer ID: 00-7D, 7E=Universal non-realtime, 7F=Universal realtime
 *  # Device number: 00-0F = dev 1 -16, 7F=all
 *  # Model ID: 00-7F
 *  # 0 or more data bytes
 *  # Checksum: 00-7F (all data + checksum = 0 for lowest 7 bits)
 *  # End of Exclusive: F7 
 *
 * eg: 
 * F0 7F 7F 01 01 20 00 03 01 F7
 * F0 7F 7F 06 44 06 01 20 00 03 01 00 F7
 * timecode 00:00:03:01 (@ 25fps) -- Roland MTC
 *
 *  roland MTC sysex real time message - reverse engineered format
 *         01 01 [fps: bit 6..5 hour: bit4..0] [min] [sec] [frame] F7
 *         06 xx 06 xx [fps (bit 6..5) hour: bit4..0)] [min] [sec] [frame] Checksum F7
 *
 *  fps/hour:  
 *    (011HHHHH):  30 fps, 29fps - non drop
 *    (010HHHHH :  29 fps (drop)
 *    (001HHHHH):  25 fps
 *    (000HHHHH):  24 fps
 *  min,sec,frame are 6bit values
 */

void midi_send_sysex(snd_seq_event_t *ev) {
  snd_seq_ev_set_direct(ev);
  snd_seq_ev_set_source(ev, portnum);
  snd_seq_ev_set_dest(ev, SND_SEQ_ADDRESS_SUBSCRIBERS, portnum);

  snd_seq_event_output(seq, ev);
  snd_seq_drain_output(seq);
}

/** MTC Full Frame message 
 */
void send_mtc_sxA(SMPTETime *stime) {
  snd_seq_event_t ev;
  snd_seq_ev_set_fixed(&ev); /* reset */

  unsigned char sysex[10] = {
    0xf0, 0x7f, 0x7f, 0x01, 0x01, 
    0x00, 0x00, 0x00, 0x00, 0xf7
  };
  sysex[5] |= (char) 0x20; // 25fps
  sysex[5] |= (char) (stime->hours&0x1f);
  sysex[6] |= (char) (stime->mins&0x7f);
  sysex[7] |= (char) (stime->secs&0x7f);
  sysex[8] |= (char) (stime->frame&0x7f);
  snd_seq_ev_set_sysex(&ev, 10, sysex);

  midi_send_sysex(&ev);
}

/** MTC Realtime Event Play start/stop Message
 */
#if 0
void send_mtc_sxB(SMPTETime *stime) {
  snd_seq_event_t ev;
  snd_seq_ev_set_fixed(&ev); /* reset */

  unsigned char sysex[13] = {
    0xf0, 0x7f, 0x7f, 0x06, 0x44, 0x06, 0x01,
    0x00, 0x00, 0x00, 0x00, 
    0x00, 0xf7
  };
  sysex[7]  |= (char) 0x20; // 25fps
  sysex[7]  |= (char) (stime->hours&0x1f);
  sysex[8]  |= (char) (stime->mins&0x7f);
  sysex[9]  |= (char) (stime->secs&0x7f);
  sysex[10] |= (char) (stime->frame&0x7f);

  int checksum = (sysex[7] + sysex[8] + sysex[9] + sysex[10] + 0x3f)&0x7f ;
  sysex[11]  = (char) (127-checksum); //checksum
  snd_seq_ev_set_sysex(&ev, 13, sysex);
  midi_send_sysex(&ev);
}
#endif

void send_mtc_sx(SMPTETime *stime) {
  send_mtc_sxA(stime);
//send_mtc_sxB(stime);
}

// ------------------------

int next_quarter_frame_to_send = 0;

int send_midi_timecode(SMPTETime *stime) {
  unsigned char mtc_msg=0;
  int mtc_tc = 0x20; // 25 fps
  switch(next_quarter_frame_to_send) {
    case 0: mtc_msg =  0x00 |  (stime->frame&0xf); break;
    case 1: mtc_msg =  0x10 | ((stime->frame&0xf0)>>4); break;
    case 2: mtc_msg =  0x20 |  (stime->secs&0xf); break;
    case 3: mtc_msg =  0x30 | ((stime->secs&0xf0)>>4); break;
    case 4: mtc_msg =  0x40 |  (stime->mins&0xf); break;
    case 5: mtc_msg =  0x50 | ((stime->mins&0xf0)>>4); break;
    case 6: mtc_msg =  0x60 |  ((mtc_tc|stime->hours)&0xf); break;
    case 7: mtc_msg =  0x70 | (((mtc_tc|stime->hours)&0xf0)>>4); break;
  }

  snd_seq_event_t ev;
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, portnum);
  snd_seq_ev_set_dest(&ev, SND_SEQ_ADDRESS_SUBSCRIBERS, portnum);
  ev.queue = SND_SEQ_QUEUE_DIRECT;
  ev.type = SND_SEQ_EVENT_QFRAME;
  ev.data.control.value= mtc_msg;

  snd_seq_event_output(seq, &ev);
  snd_seq_drain_output(seq);

  next_quarter_frame_to_send++;
  if (next_quarter_frame_to_send >= 8) {
    next_quarter_frame_to_send = 0;
    return 1;
  }
  return 0;
}


int midi_connected(void) {
  if (seq) return (1);
  return (0);
}

void midi_close(void) {
  if(!seq) return;
  if (want_verbose) printf("closing alsa midi...");
  snd_seq_close(seq);
  seq=NULL;
}

void midi_open(char *port_name) {
  int err=0;
  snd_seq_addr_t port;
  char seq_name[32];
  snprintf(seq_name,32,"ltc2amidi");

  if (seq) return; // already open

  /* open sequencer */ // SND_SEQ_OPEN_INPUT
  if ((err = snd_seq_open(&seq, "default", SND_SEQ_OPEN_OUTPUT, 0)) <0 ) {
    fprintf(stderr,"cannot open alsa sequencer: %s\n", snd_strerror(err));
    seq=NULL;
    return;
  }

  if ((err = snd_seq_set_client_name(seq, seq_name)) <0 ) {
    fprintf(stderr,"cannot set client name: %s\n", snd_strerror(err));
    midi_close();
    return;
  }

  if ((portnum = snd_seq_create_simple_port(seq, "MTC out", 
      SND_SEQ_PORT_CAP_READ | SND_SEQ_PORT_CAP_SUBS_READ, 
      SND_SEQ_PORT_TYPE_MIDI_GENERIC | SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
    fprintf(stderr,"cannot create midi port: %s\n", snd_strerror(portnum));
    midi_close();
    return;
  }

  if (port_name) {
    err = snd_seq_parse_address(seq, &port, port_name);
    if (err < 0) {
      fprintf(stderr,"Cannot find port %s - %s\n", port_name, snd_strerror(err));
    }
    err = snd_seq_connect_to(seq, 0, port.client, port.port);
    if (err < 0) {
      fprintf(stderr,"Cannot connect to port %d:%d - %s\n", port.client, port.port, snd_strerror(err));
    }
  }

  snd_seq_nonblock(seq, 1);
}

/* vi:set ts=8 sts=2 sw=2: */
