/* jack linear time code en/decoder
    Copyright (C) 2006 Robin Gareus
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include <signal.h>
#include <getopt.h>
#include <jack/jack.h>
#include <ltcsmpte.h>
#include <pthread.h>

#include <config.h>

/* options - TODO */
int nports = 1; //< total number of jack ports to use.
int want_verbose = 0;
char *midi_port_connect = NULL;
int fps_num = 25;
int fps_den = 1;

/* ltc2amidi.c prototypes */
int midi_connected(void);
void midi_open(char *midiid);
void midi_close(void);
void send_mtc_sx(SMPTETime *stime);

/* global vars */
jack_port_t **input_port;
jack_default_audio_sample_t **in;
jack_nframes_t *latency;
jack_client_t *client;
uint32_t j_samplerate;

SMPTEDecoder **decoder;
long int monotonic_fcnt = 0;

pthread_t thread_id;
pthread_mutex_t ltc_thread_lock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t  data_ready = PTHREAD_COND_INITIALIZER;

/**
 * cleanup and exit
 * call this function only _after_ everything has been initialized!
 */
void cleanup(int sig) {
  int i;

  jack_deactivate(client);
  midi_close();
  jack_client_close (client);

  for (i = 0; i < nports; i++) {
    SMPTEFreeDecoder(decoder[i]);
  }
  free(in);
  free(input_port);
  free(decoder);
  printf("bye.\n");
  exit(0);
}

#ifdef FRAME_FRACTIONS
 // libltcsmpte does not [yet] export decodeBitCnt
#include "/home/rgareus/data/coding/ltcsmpte/libltcsmpte/trunk/src/decoder.h"
#endif
/**
 * 
 */
int myProcess(SMPTEDecoder *d, int chn) {
  SMPTEFrameExt frame;
  int errors;
  int i=0;
  int rv=0;
  int flush = 0; // XXX process only last LTC (0: all in decoder queue)
  while (flush && SMPTEDecoderRead(d,&frame)) {i++;}
  while (i || SMPTEDecoderRead(d,&frame)) {
    SMPTETime stime;
    i=0;
    
    // TODO: use frame.posinfo for latency compensation..
    SMPTEFrameToTime(&frame.base,&stime);
    SMPTEDecoderErrors(d,&errors);

    //send_mtc(chn,&stime);
    send_mtc_sx(&stime);

    rv|=1<<chn;
#ifdef DEBUG
    int ms;
    SMPTEDecoderFrameToMillisecs(d,&frame,&ms);
    printf("Ch%i: %02d:%02d:%02d:%02d %8d %d \n",chn+1,
      stime.hours,stime.mins,
      stime.secs,stime.frame,
      ms,
      errors);		
#endif
  }
#ifdef FRAME_FRACTIONS
  if (rv==0) {
#ifdef DEBUG
     printf("LTC frame not finished: %i/80\n",d->decodeBitCnt);
#endif
  }
#endif

  return rv;
}

/**
 *
 */
int decodeltc(int nports, jack_nframes_t nframes, jack_default_audio_sample_t **in, jack_nframes_t *latencies, long int posinfo) {
  unsigned char sound[4096];
  jack_default_audio_sample_t val;
  int chn;
  size_t i;
 
  assert(nframes <= 4096);
 
  for (chn = 0; chn < nports; chn++) {
    for (i = 0; i < nframes; i++) {
      int snd;
      val=*(in[chn]+i);
      snd=(int)rint((127.0*val)+128.0);
      sound[i] = (unsigned char) (snd&0xff);
    }
    SMPTEDecoderWrite(decoder[chn],sound,nframes,posinfo-latencies[i]);
  }
  return 0;
}


/**********************************************************
 * JACK CLIENT CODE
 */

/**
 * jack audio process callback
 */
int process (jack_nframes_t nframes, void *arg) {
  int i;

  for (i=0;i<nports;i++) {
    in[i] = jack_port_get_buffer (input_port[i], nframes);
    latency[i] = jack_port_get_total_latency(client,input_port[i]);
  }

  decodeltc(nports, nframes, in, latency, monotonic_fcnt);

  monotonic_fcnt += nframes; // XXX we don't need that here - TODO simply use latency as posinfo and subtract it later..

  if (pthread_mutex_trylock (&ltc_thread_lock) == 0) {
    pthread_cond_signal (&data_ready);
    pthread_mutex_unlock (&ltc_thread_lock);
  }

  return 0;      
}

/**
 * JACK calls this shutdown_callback if the server ever shuts down or
 * decides to disconnect the client.
 */
void jack_shutdown (void *arg) {
  fprintf(stderr,"recv. shutdown request from jackd.\n");
  cleanup(0);
}

/**
 * open a client connection to the JACK server 
 */
void init_jack(const char *client_name, const char *server_name, jack_options_t options) {
  jack_status_t status;
  client = jack_client_open (client_name, options, &status, server_name);
  if (client == NULL) {
    fprintf (stderr, "jack_client_open() failed, status = 0x%2.0x\n", status);
    if (status & JackServerFailed) {
      fprintf (stderr, "Unable to connect to JACK server\n");
    }
    exit(1);
  }
  if (status & JackServerStarted) {
    fprintf (stderr, "JACK server started\n");
  }
  if (status & JackNameNotUnique) {
    client_name = jack_get_client_name(client);
    fprintf (stderr, "unique name `%s' assigned\n", client_name);
  }
  jack_set_process_callback (client, process, 0);
  jack_on_shutdown (client, jack_shutdown, 0);
  j_samplerate=jack_get_sample_rate (client);
}

/**
 *
 */
void jack_portsetup(void) {
  int i;
  input_port = (jack_port_t **) malloc (sizeof (jack_port_t *) * nports);
  in = (jack_default_audio_sample_t **) calloc (nports,sizeof (jack_default_audio_sample_t *));
  latency = (jack_nframes_t *) calloc (nports,sizeof (jack_nframes_t));
  decoder= (SMPTEDecoder **) calloc (nports,sizeof (SMPTEDecoder *));
  
  for (i = 0; i < nports; i++) {
    FrameRate *fps;
    decoder[i] = SMPTEDecoderCreate(j_samplerate,(fps=FR_create(fps_num,fps_den,FRF_NONE)),8,1);
    FR_free(fps);
  }
  
  for (i = 0; i < nports; i++) {
    char name[64];
    sprintf (name, "input%d", i+1);
    if ((input_port[i] = jack_port_register (client, name, JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0)) == 0) {
      fprintf (stderr, "cannot register input port \"%s\"!\n", name);
      jack_client_close (client);
      exit (1);
    }
  }
}

/**
 *
 */
void main_loop(void) {
  int chn;
  pthread_mutex_lock (&ltc_thread_lock);
  while (1) {
    for (chn = 0; chn < nports; chn++) { 
      myProcess(decoder[chn],chn);
    } /* foreach channel */
    pthread_cond_wait (&data_ready, &ltc_thread_lock);
  } /* while running */
  pthread_mutex_unlock (&ltc_thread_lock);
}

/**************************
 * main application code 
 */

static void usage (int status);
static int decode_switches (int argc, char **argv);

/* The name the program was run with, stripped of any leading path. */
char *program_name;

/* Option flags and variables */

static struct option const long_options[] =
{
  {"help", no_argument, 0, 'h'},
  {"version", no_argument, 0, 'V'},
  {NULL, 0, NULL, 0}
};

int main (int argc, char **argv) {
  int i;
  const char *client_name;
  const char *server_name = NULL;
  jack_options_t options = JackNullOption;

  program_name = argv[0];

  i = decode_switches (argc, argv);

  client_name = strrchr(argv[0], '/');
  if (client_name == 0) {
    client_name = argv[0];
  } else {
    client_name++;
  }

  init_jack(client_name, server_name, options);
  jack_portsetup();

 // TODO: autoconnect JACK port option 

  if (want_verbose) 
    printf ("engine sample rate: %" PRIu32 "\n", j_samplerate);

  midi_open(midi_port_connect); // TODO option for autoconnect MIDI port

  if (!midi_connected()) {
    fprintf (stderr, "cannot open midi port.\n");
    cleanup(0);
  }

  if (jack_activate (client)) {
    fprintf (stderr, "cannot activate client.\n");
    cleanup(0);
  }

  main_loop();

  cleanup(0);
  return(0);
}

/**
 * Set all the option flags according to the switches specified.
 * Return the index of the first non-option argument.  
 */
static int decode_switches (int argc, char **argv) {
  int c;


  while ((c = getopt_long (argc, argv, 
			   "h"	/* help */
			   "V",	/* version */
			   long_options, (int *) 0)) != EOF)
    {
      switch (c)
	{
	case 'V':
	  printf ("jltc version %s\n", VERSION);
	  printf (" (c) GPL 2007 Robin Gareus  http://ltcsmpte.sf.net/\n");
	  exit (0);

	case 'h':
	  usage (0);

	default:
	  usage (EXIT_FAILURE);
	}
    }

  return optind;
}

static void usage (int status) {
  printf ("%s - JACK app to parse/generate linear time code\n", program_name);
  printf ("Usage: %s [OPTION]\n", program_name);
  printf (" Options:\n\
  -h, --help                 display this help and exit\n\
  -V, --version              output version information and exit\n\
");
  exit (status);
}

/* vi:set ts=8 sts=2 sw=2: */
